from __future__ import annotations

from bavisitter.model.message_model import ChunkType


def get_chunk_type(chunk, previous_message=None) -> ChunkType | None:
  if "start" in chunk and chunk["start"]:
    return "start"
  if "end" in chunk and chunk["end"]:
    return "end"
  if "content" in chunk and isinstance(chunk["content"], str):
    if (
      previous_message
      and chunk["role"] == "assistant"
      and "```" in (previous_message["content"] + chunk["content"])
    ):
      if previous_message["type"] == "message":
        return "code_start"
      elif (
        previous_message["type"] == "code"
        and "```" in previous_message["content"]
      ):
        return "continue"
      else:
        return "code_end"

    return "continue"
  return None



from .deep_equal import deep_equal
from .file_browser import load_artifact, save_artifact
from .get_chunk_type import get_chunk_type
from .set_interpreter import set_interpreter
from .system_prompt import SYSTEM_PROMPT

__all__ = [
  "SYSTEM_PROMPT",
  "get_chunk_type",
  "set_interpreter",
  "load_artifact",
  "save_artifact",
  "deep_equal",
]



from __future__ import annotations

from typing import Any, Literal, TypedDict

IPCType = Literal["reqest", "resonse"]


class IPCModel(TypedDict):
  type: IPCType
  content: Any
  endpoint: str
  uuid: str



from .ipc_model import IPCModel
from .message_model import (
  ChunkType,
  MessageModel,
  MessageType,
  RoleType,
  StreamChunkModel,
)

__all__ = [
  "MessageModel",
  "MessageType",
  "RoleType",
  "StreamChunkModel",
  "ChunkType",
  "IPCModel",
]



def deep_equal(a, b):
  if a == b:
    return True
  if type(a) != type(b):
    return False
  if isinstance(a, dict):
    if len(a) != len(b):
      return False
    for key in a:
      if key not in b:
        return False
      if not deep_equal(a[key], b[key]):
        return False
    return True
  if isinstance(a, list):
    if len(a) != len(b):
      return False
    for i in range(len(a)):
      if not deep_equal(a[i], b[i]):
        return False
    return True
  return False


__all__ = [
  "deep_equal",
]



from __future__ import annotations

import pathlib

import pandas as pd


def load_artifact(path: str):
  if not pathlib.Path(path).exists():
    return None

  # check if file type is supported by pandas dataframe
  if path.endswith((".csv", ".json", ".xlsx", ".html", ".parquet", ".feather")):
    file_extension = pathlib.Path(path).suffix
    read_method = f"read_{file_extension[1:]}"
    df: pd.DataFrame = getattr(pd, read_method)(path)
    df = df.where(pd.notnull(df), None)
    return df.to_json(orient="records")

  return None


def save_artifact(record: list[dict], path: str):
  df = pd.DataFrame(record)
  df.to_csv(path, index=False)


__all__ = ["load_artifact"]



from typing import TYPE_CHECKING, Literal

from bavisitter.utils.system_prompt import SYSTEM_PROMPT

if TYPE_CHECKING:
  from interpreter import OpenInterpreter

PRELOAD_SCRIPT = [
  "import pandas as pd\ndel pd.DataFrame._repr_html_",
  "df = pd.read_csv('artifacts/data.csv')",
]


def set_interpreter(
  interpreter: "OpenInterpreter",
  model: str = "gpt-4-turbo",
  safe_mode: Literal["off", "auto"] = "auto",
  auto_run: bool = True,
  system_prompt: str = SYSTEM_PROMPT,
):
  interpreter.system_message = system_prompt
  interpreter.llm.model = model
  interpreter.safe_mode = safe_mode
  interpreter.auto_run = auto_run
  interpreter.llm.max_tokens = 4096  # type: ignore
  interpreter.llm.context_window = 128_000  # type: ignore
  for script in PRELOAD_SCRIPT:
    list(interpreter.computer.terminal.run("python", script))



from __future__ import annotations

import importlib.metadata
import pathlib
import re
from copy import copy
from typing import Literal

import anywidget
import pandas as pd
from interpreter import OpenInterpreter
from traitlets import (
  Bool,
  Dict,
  HasTraits,
  List,
  TraitError,
  Unicode,
  default,
  observe,
  validate,
)

from bavisitter.model import (
  IPCModel,
  MessageModel,
  MessageType,
  RoleType,
  StreamChunkModel,
)
from bavisitter.utils import (
  deep_equal,
  get_chunk_type,
  load_artifact,
  save_artifact,
  set_interpreter,
)

try:
  __version__ = importlib.metadata.version("bavisitter")

except importlib.metadata.PackageNotFoundError:
  __version__ = "unknown"

match = re.compile(r"```(\w+)\n")


class Bavisitter(anywidget.AnyWidget, HasTraits):
  _esm = pathlib.Path(__file__).parent / "static" / "main.js"
  messages = List(Dict(value_trait=Unicode(), key_trait=Unicode())).tag(
    sync=True
  )
  streaming = Bool(default_value=False).tag(sync=True)
  color_mode = Unicode(default_value="light").tag(sync=True)
  ipc_queue = List(Dict()).tag(sync=True)

  def __init__(
    self,
    df: pd.DataFrame,
    model: str = "gpt-4-turbo-preview",
    safe_model: Literal["auto", "off"] = "auto",
    auto_run: bool = True,
    color_mode: Literal["light", "dark"] = "light",
    artifact_path: str = "artifacts",
    use_cli: bool = False,
    **kwargs,
  ):
    super().__init__(**kwargs)

    if not pathlib.Path(artifact_path).exists():
      pathlib.Path(artifact_path).mkdir()

    df.to_csv(f"{artifact_path}/data.csv", index=False)

    # set widget properties
    self.model = model
    self.chunks = []
    self.color_mode = color_mode
    self.use_cli = use_cli
    # initialize the open interpreter
    self.interpreter = OpenInterpreter()
    set_interpreter(self.interpreter, model, safe_model, auto_run)

  @default("messages")
  def _default_messages(self):
    return []

  @default("streaming")
  def _default_stream(self):
    return False

  @default("color_mode")
  def _default_color_mode(self):
    return "light"

  @default("ipc_queue")
  def _default_ipc_queue(self):
    return []

  @validate("color_mode")
  def _validate_color_mode(self, proposal):
    if proposal["value"] not in ["light", "dark"]:
      raise TraitError("Invalid color mode. Must be 'light' or 'dark'")

    return proposal["value"]

  @validate("ipc_queue")
  def _validate_ipc_queue(self, proposal):
    for message in proposal["value"]:
      if "type" not in message:
        raise TraitError("Invalid message: 'type' key is missing")

    return proposal["value"]

  @validate("messages")
  def _validate_messages(self, proposal):
    try:
      for message in proposal["value"]:
        MessageModel(**message)
    except Exception as e:
      raise TraitError(f"Invalid message: {e}")

    return proposal["value"]

  @observe("ipc_queue")
  def handle_ipc_queue(
    self, change: dict[Literal["new", "old"], list[IPCModel]]
  ):
    if len(change["new"]) and change["new"][-1]["type"] == "request":
      if change["new"][-1]["endpoint"] == "load_artifact":
        try:
          data = load_artifact(change["new"][-1]["content"])
          self.ipc_queue = [
            *self.ipc_queue,
            {
              "type": "response",
              "content": data,
              "endpoint": change["new"][-1]["endpoint"],
              "uuid": change["new"][-1]["uuid"],
            },
          ]
        except Exception:
          self.ipc_queue = [
            *self.ipc_queue,
            {
              "type": "response",
              "content": None,
              "endpoint": change["new"][-1]["endpoint"],
              "uuid": change["new"][-1]["uuid"],
            },
          ]

      if change["new"][-1]["endpoint"] == "save_artifact":
        try:
          path = change["new"][-1]["content"]["path"]
          records = change["new"][-1]["content"]["records"]
          save_artifact(records, path)
          self.ipc_queue = [
            *self.ipc_queue,
            {
              "type": "response",
              "content": "success",
              "endpoint": change["new"][-1]["endpoint"],
              "uuid": change["new"][-1]["uuid"],
            },
          ]
        except Exception:
          self.ipc_queue = [
            *self.ipc_queue,
            {
              "type": "response",
              "content": None,
              "endpoint": change["new"][-1]["endpoint"],
              "uuid": change["new"][-1]["uuid"],
            },
          ]

      else:
        self.ipc_queue = [
          *self.ipc_queue,
          {
            "type": "response",
            "content": None,
            "endpoint": change["new"][-1]["endpoint"],
            "uuid": change["new"][-1]["uuid"],
          },
        ]

  @observe("messages")
  def handle_user_chat(self, change):
    if self.streaming:
      return

    if deep_equal(change["new"], change["old"]):
      return

    if len(change["new"]) > 0 and change["new"][-1]["role"] == "user":
      self.streaming = True
      chats = self.interpreter.chat(
        change["new"][-1]["content"], display=False, stream=True
      )
      if chats is None:
        self.streaming = False
        return
      for chunk in chats:
        self.chunks.append(chunk)
        if (
          "content" in chunk
          and isinstance(chunk["content"], str)
          and "disabled or not supported." in chunk["content"]
        ):
          self.interpreter.messages = self.interpreter.messages[:-1]
          continue
        self.handle_chunk(chunk)  # type: ignore
      self.streaming = False

    if len(change["new"]) == 0 or (
      len(change["new"]) > 0 and change["new"][-1]["role"] == "assistant"
    ):
      self.load_messages(change["new"])

  def handle_chunk(self, chunk: StreamChunkModel):
    chunk_type = get_chunk_type(
      chunk, self.messages[-1] if len(self.messages) else None
    )

    if chunk_type == "start":
      format = chunk["format"] if "format" in chunk else None
      format = "output" if chunk["type"] == "console" else format
      self.append_message(
        role=chunk["role"], type=chunk["type"], content="", format=format
      )

    elif chunk_type == "code_start":
      self.append_message(
        role=chunk["role"], type="code", content=chunk["content"]
      )

    elif chunk_type == "continue":
      if self.use_cli:
        print(chunk["content"], end="")

      if (
        self.messages[-1]["type"] == "code"
        and "```" in self.messages[-1]["content"]
        and match.match(self.messages[-1]["content"])
      ):
        content_match = match.match(self.messages[-1]["content"])
        format = content_match.group(1) if content_match else ""
        code = re.sub(r"```(\w+)\n", "", self.messages[-1]["content"])
        self.messages = [
          *self.messages[:-1],
          {
            "role": "assistant",
            "type": "code",
            "content": code + chunk["content"],
            "format": format,
          },
        ]
      else:
        self.append_content(chunk["content"])

    elif chunk_type == "end":
      if self.messages[-1]["content"] == "":
        self.messages = [*self.messages[:-1]]

    elif chunk_type == "code_end":
      content: str = self.messages[-1]["content"] + chunk["content"]
      code = content.split("```")[0]
      self.messages = [
        *self.messages[:-1],
        {
          "role": "assistant",
          "type": "code",
          "content": code,
          "format": self.messages[-1]["format"],
        },
        {
          "role": "assistant",
          "type": "message",
          "content": "",
        },
      ]

  def append_content(self, content: str):
    new_message = copy(self.messages[-1])
    new_message["content"] += content
    self.messages = [*self.messages[:-1], new_message]

  def append_message(
    self,
    role: RoleType = "assistant",
    content: str = "",
    format: str | None = None,
    type: MessageType = "message",
  ):
    new_message = {
      "role": role,
      "type": type,
      "content": content,
    }

    if format:
      new_message["format"] = format

    self.messages = [
      *self.messages,
      new_message,
    ]

  def load_messages(self, messages):
    self.messages = messages
    self.interpreter.messages = messages



SYSTEM_PROMPT = """You are a Visualization Assistant, a world-class data analyst that can complete any goal.

First, write a plan for designing vega-lite visualization. **Always recap the plan between each code block** (you have extreme short-term memory loss, so you need to recap the plan between each message block to retain it).

When you execute code, it will be executed **on the user's machine**. The user has given you **full and complete permission** to execute any code necessary to complete the task. 

In general, try to **make plans** with as few steps as possible.

Write messages to the user in Markdown.

As for actually executing code to carry out that plan, for *stateful* languages (like Python, javascript, and shell, but NOT for HTML, which starts from 0 every time) **it's critical not to try to do everything in one code block.**

You should try something, print information about it, then continue from there in tiny, informed steps. You will never get it on the first try, and attempting it in one go will often lead to errors you can't see.

You can install new packages with pip and continue working.

You are capable of **any** task.

The visualization **must** be written with vega-lite in JSON code blocks like the one below. Never execute vega-lite code; only write it in JSON code blocks.


```json
{
    "$schema": "https://vega.github.io/schema/vega-lite/v5.json",
    "data": {"url": "artifacts/data.csv"},
    "mark": "bar",
    "encoding": {
        "x": {"field": "a", "type": "ordinal"},
        ...
}
```

Never use other visualization libraries like matplotlib, seaborn, plotly, etc. Only use vega-lite.


Carefully analyze and visualize the data. The data is already loaded in enviorinment as a variable named `df`.

If you write vega-lite specification, then I will render it for visualization.

If you revised the data, you should save it in the `artifacts` directory with a distinct name (e.g., `artifacts/data_foo_bar.csv`). Do not overwrite files in the `artifacts` directory.

The user's need is as follows:

""".strip()
__all__ = ["SYSTEM_PROMPT"]



from __future__ import annotations

from typing import Literal, Optional, TypedDict, Union

AssistantRoleType = Literal["assistant"]
UserRoleType = Literal["user"]
SystemRoleType = Literal["computer"]
MessageType = Literal["message", "code", "console"]
ChunkType = Literal["start", "continue", "end", "code_start", "code_end"]

RoleType = Union[AssistantRoleType, UserRoleType, SystemRoleType]


class MessageModel(TypedDict):
  role: RoleType
  type: MessageType
  format: str | None
  content: str


class StreamChunkModel(MessageModel, total=False):
  start: Optional[bool]
  end: Optional[bool]
